""" Admin Plugin, Can manage your Group. """

# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import ClassVar, Optional

from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_type import ChatType
from pyrogram.errors import (
    BotChannelsNa,
    ChatAdminRequired,
    FloodWait,
    UserAdminInvalid,
    UserCreator,
    UserIdInvalid,
    UserPrivacyRestricted,
)
from pyrogram.types import Chat, ChatPrivileges, User

from anjani import command, filters, plugin, util


class Admins(plugin.Plugin):
    name: ClassVar[str] = "Admins"
    helpable: ClassVar[bool] = True

    @command.filters(filters.can_pin)
    async def cmd_pin(self, ctx: command.Context) -> Optional[str]:
        """Pin message on chats"""
        if not ctx.msg.reply_to_message:
            return await self.text(ctx.chat.id, "error-reply-to-message")

        is_silent = True
        if ctx.input and ctx.input in {
            "notify",
            "loud",
            "violence",
        }:
            is_silent = False

        await ctx.msg.reply_to_message.pin(disable_notification=is_silent)
        return None

    @command.filters(filters.can_pin)
    async def cmd_unpin(self, ctx: command.Context) -> Optional[str]:
        """Unpin message on chats"""
        chat = ctx.chat

        if ctx.input and ctx.input == "all":
            await self.bot.client.unpin_all_chat_messages(chat.id)
        elif not ctx.msg.reply_to_message:
            chat = await self.bot.client.get_chat(chat.id)
            if not isinstance(chat, Chat):
                raise ValueError("Invalid Chat")

            if not chat.pinned_message:
                return await self.text(chat.id, "no-pinned-message")

            pinned = chat.pinned_message.id
            await self.bot.client.unpin_chat_message(chat.id, pinned)
        else:
            await ctx.msg.reply_to_message.unpin()

        return None

    @command.filters(filters.can_change_info)
    async def cmd_setgpic(self, ctx: command.Context) -> Optional[str]:
        """Set group picture"""
        msg = ctx.msg.reply_to_message or ctx.msg
        file = msg.photo or None

        if not file:
            return await self.text(ctx.chat.id, "gpic-no-photo")

        await self.bot.client.set_chat_photo(ctx.chat.id, photo=file.file_id)
        return await self.text(ctx.chat.id, "gpic-success-changed")

    async def cmd_adminlist(self, ctx: command.Context) -> str:
        """Get list of chat admins"""
        chat = ctx.chat
        if chat.type == ChatType.PRIVATE:
            return await self.text(chat.id, "err-chat-groups")
        admins = ""

        async for admin in util.tg.get_chat_admins(ctx.bot.client, chat.id):
            if admin.status == ChatMemberStatus.OWNER:
                admins += f"• {admin.user.mention} (**Creator**)\n"
            elif admin.user.id == self.bot.uid:
                admins += f"• {admin.user.mention} (**Me**)\n"
            elif admin.user.id == ctx.author.id:
                admins += f"• {admin.user.mention} (**You**)\n"
            else:
                admins += f"• {admin.user.mention}\n"

        return admins

    @command.filters(filters.can_restrict)
    async def cmd_zombies(self, ctx: command.Context) -> str:
        """Kick all deleted acc in group."""
        chat = ctx.chat
        zombie = 0

        await ctx.respond(await self.text(chat.id, "finding-zombie"))
        async for member in self.bot.client.get_chat_members(chat.id):  # type: ignore
            if member.user.is_deleted:
                zombie += 1
                try:
                    await self.bot.client.ban_chat_member(chat.id, member.user.id)
                except UserAdminInvalid:
                    zombie -= 1
                except FloodWait as flood:
                    await asyncio.sleep(flood.value)  # type: ignore

        if zombie == 0:
            return await self.text(chat.id, "zombie-clean")

        return await self.text(chat.id, "cleaning-zombie", zombie)

    @command.filters(filters.can_promote)
    async def cmd_promote(self, ctx: command.Context, user: Optional[User] = None) -> Optional[str]:
        """Bot promote member, required Both permission of can_promote"""
        chat = ctx.chat
        if not chat:
            return

        if not user:
            if ctx.input:
                return await self.text(chat.id, "err-peer-invalid")

            if not ctx.msg.reply_to_message or not ctx.msg.reply_to_message.from_user:
                return await self.text(chat.id, "no-promote-user")

            user = ctx.msg.reply_to_message.from_user

        if user.id == ctx.author.id:
            return await self.text(chat.id, "promote-error-self")

        if user.id == self.bot.uid:
            return await self.text(chat.id, "error-its-myself")

        bot, _ = await util.tg.fetch_permissions(self.bot.client, chat.id, user.id)
        if not bot:
            return await self.text(chat.id, "promote-error-perm")

        try:
            await chat.promote_member(
                user_id=user.id, privileges=bot.privileges, title="Administrator"
            )
        except ChatAdminRequired:
            return await self.text(chat.id, "promote-error-perm")
        except UserIdInvalid:
            return await self.text(chat.id, "promote-error-invalid")
        except UserPrivacyRestricted:
            return await self.text(chat.id, "promote-error-privacy-restricted")

        return await self.text(chat.id, "promote-success")

    @command.filters(filters.can_promote)
    async def cmd_demote(self, ctx: command.Context, user: Optional[User] = None) -> Optional[str]:
        """Demoter Just owner and promoter can demote admin."""
        chat = ctx.chat
        if not chat:
            return

        if not user:
            if ctx.input:
                return await self.text(chat.id, "err-peer-invalid")

            if not ctx.msg.reply_to_message or not ctx.msg.reply_to_message.from_user:
                return await self.text(chat.id, "no-demote-user")

            user = ctx.msg.reply_to_message.from_user

        if user.id == ctx.author.id:
            return await self.text(chat.id, "demote-error-self")

        if user.id == self.bot.uid:
            return await self.text(chat.id, "error-its-myself")

        try:
            await chat.promote_member(
                user_id=user.id,
                title="",
                privileges=ChatPrivileges(
                    can_manage_chat=False,
                    can_delete_messages=False,
                    can_manage_video_chats=False,
                    can_restrict_members=False,
                    can_promote_members=False,
                    can_change_info=False,
                    can_post_messages=False,
                    can_edit_messages=False,
                    can_invite_users=False,
                    can_pin_messages=False,
                    is_anonymous=False,
                ),
            )
        except (BotChannelsNa, ChatAdminRequired):
            return await self.text(chat.id, "demote-error-perm")
        except UserCreator:
            return await self.text(chat.id, "demote-error-creator")

        return await self.text(chat.id, "demote-success")



from prometheus_client import Counter, Gauge

EventCount = Counter(
    "anjani_event_count",
    "Number of events being processed",
    labelnames=["type"],
)
SpamPredictionStat = Counter(
    "anjani_spam_prediction_stat",
    "Number of spam prediction event",
    labelnames=["status"],
)
MessageStat = Counter(
    "anjani_message_stat",
    "Number of message",
    labelnames=["type"],
)
CommandCount = Counter("anjani_command_stats", "Number of coomand", labelnames=["name"])
UnhandledError = Counter("anjani_unhandled_error", "Number of unhandled error", labelnames=["type"])

EventLatencySecond = Gauge(
    "anjani_event_latency",
    "Latency of event processed",
    labelnames=["type"],
    unit="second",
)
CommandLatencySecond = Gauge(
    "anjani_command_latency",
    "Latency of command processed",
    labelnames=["name"],
    unit="second",
)



"""Bot stats plugin"""
# Copyright (C) 2020 - 2022  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import ClassVar, Optional

from pyrogram.types import CallbackQuery, InlineKeyboardButton, InlineKeyboardMarkup

from anjani import command, filters, listener, plugin, util


class Topics(plugin.Plugin):
    name: ClassVar[str] = "Topic"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("CHATS")

    @listener.filters(filters.regex(r"topic_action_(.*)"))
    async def on_callback_query(self, query: CallbackQuery) -> None:
        action = query.matches[0].group(1)
        chat = query.message.chat

        user = await chat.get_member(query.from_user.id)
        if not user.privileges or not user.privileges.can_manage_topics:
            await query.answer(await self.text(chat.id, "error-no-rights"))
            return

        if action == "cancel":
            await query.message.delete()
            return
        if action == "remove":
            await self.bot.client.delete_forum_topic(chat.id, query.message.message_thread_id)
            return

    @command.filters(filters.can_manage_topic, aliases=["setdefaulttopic"])
    async def cmd_setactiontopic(self, ctx: command.Context) -> Optional[str]:
        """Set action topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        await self.db.update_one(
            {"chat_id": ctx.chat.id},
            {"$set": {"action_topic": ctx.msg.message_thread_id}},
            upsert=True,
        )
        return await self.text(ctx.chat.id, "topic-set")

    @command.filters(filters.can_manage_topic)
    async def cmd_actiontopic(self, ctx: command.Context) -> Optional[str]:
        """Get action topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        chat_id = ctx.chat.id
        chat = await self.db.find_one({"chat_id": ctx.chat.id})
        if not chat or not chat.get("action_topic"):
            return await self.text(ctx.chat.id, "topic-action-general")

        return await self.text(
            ctx.chat.id,
            "topic-action-custom",
            f"t.me/c/{str(chat_id).replace('-100', '')}/{chat['action_topic']}",
        )

    @command.filters(filters.can_manage_topic, aliases=["newtopic"])
    async def cmd_createtopic(self, ctx: command.Context, name: Optional[str]) -> Optional[str]:
        """create topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        if not name:
            return await self.text(ctx.chat.id, "topic-name-missing")

        await self.bot.client.create_forum_topic(ctx.chat.id, name)
        return await self.text(ctx.chat.id, "topic-created", name)

    @command.filters(filters.can_manage_topic)
    async def cmd_renametopic(self, ctx: command.Context, name: Optional[str]) -> Optional[str]:
        """Remove topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        if not name:
            return await self.text(ctx.chat.id, "topic-name-missing")

        await self.bot.client.edit_forum_topic(ctx.chat.id, ctx.msg.message_thread_id, name)
        return await self.text(ctx.chat.id, "topic-renamed", name)

    @command.filters(filters.can_manage_topic)
    async def cmd_opentopic(self, ctx: command.Context) -> Optional[str]:
        """Open topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        await self.bot.client.reopen_forum_topic(ctx.chat.id, ctx.msg.message_thread_id)
        return await self.text(ctx.chat.id, "topic-reopened", delete_after=3)

    @command.filters(filters.can_manage_topic)
    async def cmd_closetopic(self, ctx: command.Context) -> Optional[str]:
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        await ctx.respond(await self.text(ctx.chat.id, "topic-closing"), delete_after=3)
        await self.bot.client.close_forum_topic(ctx.chat.id, ctx.msg.message_thread_id)

    @command.filters(filters.can_manage_topic, aliases=["removetopic"])
    async def cmd_deletetopic(self, ctx: command.Context) -> Optional[str]:
        """Remove topic"""
        if not ctx.chat.is_forum:
            return await self.text(ctx.chat.id, "topic-non-topic")

        await ctx.respond(
            await self.text(ctx.chat.id, "topic-remove-confirm"),
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            text="⚠️ Delete",
                            callback_data="topic_action_remove",
                        )
                    ],
                    [
                        InlineKeyboardButton(
                            text="Cancel",
                            callback_data="topic_action_cancel",
                        )
                    ],
                ]
            ),
        )



"""Anjani plugin init"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import pkgutil
from pathlib import Path

current_dir = str(Path(__file__).parent)
subplugins = [
    importlib.import_module("." + info.name, __name__)
    for info in pkgutil.iter_modules([current_dir])
]

try:
    _reload_flag: bool

    # noinspection PyUnboundLocalVariable
    if _reload_flag:  # skipcq: PYL-E0601
        # Plugin has been reloaded, reload our subplugins
        for plugin in subplugins:
            importlib.reload(plugin)
except NameError:
    _reload_flag = True



"""Notes"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Any, Callable, ClassVar, Coroutine, MutableMapping, Optional

from pyrogram.enums.chat_action import ChatAction
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.errors import MediaEmpty, MessageEmpty
from pyrogram.types import Message

from anjani import command, filters, listener, plugin, util
from anjani.util.tg import Types, build_button, get_message_info, revert_button


class Notes(plugin.Plugin):
    name: ClassVar[str] = "Notes"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    ACTION: MutableMapping[int, ChatAction]
    SEND: MutableMapping[int, Callable[..., Coroutine[Any, Any, Optional[Message]]]]

    async def on_load(self):
        self.db = self.bot.db.get_collection("NOTES")
        self.ACTION = {
            Types.TEXT.value: ChatAction.TYPING,
            Types.BUTTON_TEXT.value: ChatAction.TYPING,
            Types.DOCUMENT.value: ChatAction.UPLOAD_DOCUMENT,
            Types.PHOTO.value: ChatAction.UPLOAD_PHOTO,
            Types.VIDEO.value: ChatAction.UPLOAD_VIDEO,
            Types.STICKER.value: ChatAction.CHOOSE_STICKER,
            Types.AUDIO.value: ChatAction.UPLOAD_AUDIO,
            Types.VOICE.value: ChatAction.UPLOAD_AUDIO,
            Types.VIDEO_NOTE.value: ChatAction.UPLOAD_VIDEO_NOTE,
            Types.ANIMATION.value: ChatAction.UPLOAD_VIDEO,
        }
        self.SEND = {
            Types.TEXT.value: self.bot.client.send_message,
            Types.BUTTON_TEXT.value: self.bot.client.send_message,
            Types.DOCUMENT.value: self.bot.client.send_document,
            Types.PHOTO.value: self.bot.client.send_photo,
            Types.VIDEO.value: self.bot.client.send_video,
            Types.STICKER.value: self.bot.client.send_sticker,
            Types.AUDIO.value: self.bot.client.send_audio,
            Types.VOICE.value: self.bot.client.send_voice,
            Types.VIDEO_NOTE.value: self.bot.client.send_video_note,
            Types.ANIMATION.value: self.bot.client.send_animation,
        }

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        notes = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: notes} if notes else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @listener.priority(95)
    @listener.filters(filters.regex(r"^#[\w\-]+(?!\n)$") & ~filters.outgoing)
    async def on_message(self, message: Message) -> None:
        """Notes hashtag trigger."""
        trigger = message.text or message.caption
        return await self.get_note(message, trigger.lstrip("#"))

    async def get_note(self, message: Message, name: str, noformat: bool = False) -> None:
        """Get note data and send based on types."""
        chat = message.chat
        reply_to = message.reply_to_message.id if message.reply_to_message else message.id

        data = await self.db.find_one(
            {"chat_id": chat.id, f"notes.{name}": {"$exists": True}}, {f"notes.{name}": 1}
        )
        if not data:
            return

        note = data["notes"][name]
        button = note.get("button", None)
        if noformat:
            parse_mode = ParseMode.DISABLED
            btn_text = "\n\n"
            if button:
                btn_text += revert_button(button)

            keyb = None
        else:
            parse_mode = ParseMode.MARKDOWN
            btn_text = ""
            if button:
                keyb = build_button(button)
            else:
                keyb = button

        types: int = note["type"]
        text = note["text"]
        if not text or text == "\n\n":  # Leftover from button markdown
            text = name

        await self.bot.client.send_chat_action(chat.id, self.ACTION[types])
        try:
            if types in {Types.TEXT, Types.BUTTON_TEXT}:
                await self.SEND[types](
                    chat.id,
                    text + btn_text,
                    disable_web_page_preview=True,
                    reply_to_message_id=reply_to,
                    reply_markup=keyb,
                    parse_mode=parse_mode,
                )
            elif types == Types.STICKER:
                await self.SEND[types](
                    chat.id,
                    note["content"],
                    reply_to_message_id=reply_to,
                    reply_markup=keyb,
                )
            else:
                await self.SEND[types](
                    chat.id,
                    str(note["content"]),
                    caption=text + btn_text,
                    reply_to_message_id=reply_to,
                    reply_markup=keyb,
                    parse_mode=parse_mode,
                )
        except MediaEmpty:
            await self.bot.client.send_message(
                chat.id,
                await self.get_text(chat.id, "notes-expired"),
                message_thread_id=message.message_thread_id,
            )
        except MessageEmpty:
            pass
        finally:
            await self.bot.client.send_chat_action(chat.id, ChatAction.CANCEL)

    async def cmd_get(self, ctx: command.Context) -> None:
        """Notes command trigger."""
        if not ctx.input:
            return

        if len(ctx.args) >= 2 and ctx.args[1].lower() == "noformat":
            await self.get_note(ctx.msg, ctx.args[0], noformat=True)
        else:
            await self.get_note(ctx.msg, ctx.args[0])

    @command.filters(filters.admin_only)
    async def cmd_save(self, ctx: command.Context) -> Optional[str]:
        """Save notes."""
        chat = ctx.chat
        if (
            len(ctx.args) < 2
            and not ctx.msg.reply_to_message
            or ctx.msg.reply_to_message
            and len(ctx.args) < 1
        ):
            await ctx.respond(
                await self.text(chat.id, "notes-invalid-args"), parse_mode=ParseMode.MARKDOWN
            )
            return

        trigger = ctx.args[0]
        if trigger.startswith("#") or "." in trigger or "$" in trigger:
            return await self.text(chat.id, "err-illegal-trigger")

        text, types, content, buttons = get_message_info(ctx.msg)
        _, ret = await asyncio.gather(
            self.db.update_one(
                {"chat_id": chat.id},
                {
                    "$set": {
                        "chat_name": chat.title,
                        f"notes.{trigger}": {
                            "text": text,
                            "type": types,
                            "content": content,
                            "button": buttons,
                        },
                    }
                },
                upsert=True,
            ),
            self.text(chat.id, "note-saved", trigger),
        )
        return ret

    async def cmd_notes(self, ctx: command.Context) -> str:
        """View chat notes."""
        chat = ctx.chat

        data = await self.db.find_one({"chat_id": chat.id})
        if not data or not data.get("notes"):
            return await self.text(chat.id, "no-notes")

        notes = await self.text(chat.id, "note-list", chat.title)
        for key in sorted(data["notes"].keys()):
            notes += f"`#{key}`\n"
        return notes

    @command.filters(filters.admin_only, aliases=["clear"])
    async def cmd_delnote(self, ctx: command.Context) -> str:
        """Delete chat note."""
        chat = ctx.chat
        if not ctx.input:
            return await self.text(chat.id, "notes-del-noargs")

        name = ctx.input

        data = await self.db.find_one({"chat_id": chat.id, f"notes.{name}": {"$exists": True}})
        if not data:
            return await self.text(chat.id, "no-notes")

        notes: MutableMapping[str, Any] = data["notes"]
        try:
            notes[name]
        except KeyError:
            return await self.text(chat.id, "notes-not-exist")

        _, ret = await asyncio.gather(
            self.db.update_one({"chat_id": chat.id}, {"$unset": {f"notes.{name}": ""}}),
            self.text(chat.id, "notes-deleted", name),
        )
        return ret



"""Anjani misc utils"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, Callable, Set, Tuple, Union

from pyrogram.filters import AndFilter, Filter, InvertFilter, OrFilter

from anjani.util.types import CustomFilter

if TYPE_CHECKING:
    from anjani.core import Anjani


def check_filters(filters: Union[Filter, CustomFilter], anjani: "Anjani") -> None:
    """Recursively check filters to set :obj:`~Anjani` into :obj:`~CustomFilter` if needed"""
    if isinstance(filters, (AndFilter, OrFilter, InvertFilter)):
        check_filters(filters.base, anjani)
    if isinstance(filters, (AndFilter, OrFilter)):
        check_filters(filters.other, anjani)

    # Only accepts CustomFilter instance
    if getattr(filters, "include_bot", False) and isinstance(filters, CustomFilter):
        filters.anjani = anjani


def find_prefixed_funcs(obj: Any, prefix: str) -> Set[Tuple[str, Callable[..., Any]]]:
    """Finds functions with symbol names matching the prefix on the given object."""

    results: Set[Tuple[str, Callable[..., Any]]] = set()

    for sym in dir(obj):
        if sym.startswith(prefix):
            name = sym[len(prefix) :]
            func = getattr(obj, sym)
            if not callable(func):
                continue

            results.add((name, func))

    return results


def do_nothing(*args: Any, **kwargs: Any) -> None:
    """Do nothing function"""
    return None


class StopPropagation(Exception):
    """Exception that raised to stop propagating an event"""



""" Debugging purpose """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import io
import os
import re
import sys
import traceback
from contextlib import redirect_stdout
from datetime import datetime
from html import escape
from typing import Any, ClassVar, Optional, Tuple

import pyrogram
from meval import meval
from pyrogram.enums.chat_action import ChatAction

from anjani import command, filters, plugin, util


class Debug(plugin.Plugin):
    name: ClassVar[str] = "Debug"

    async def cmd_ping(self, ctx: command.Context) -> str:
        start = datetime.now()
        await ctx.respond("Calculating response time...")
        end = datetime.now()
        latency = (end - start).microseconds / 1000

        return f"Latency: {latency} ms"

    @command.filters(filters.dev_only)
    async def cmd_eval(self, ctx: command.Context) -> Optional[str]:
        code = ctx.input
        if not code:
            return "Give me code to evaluate."

        out_buf = io.StringIO()

        async def _eval() -> Tuple[str, Optional[str]]:
            # Message sending helper for convenience
            async def send(*args: Any, **kwargs: Any) -> pyrogram.types.Message:
                return await ctx.msg.reply(*args, **kwargs)

            # Print wrapper to capture output
            # We don't override sys.stdout to avoid interfering with other output
            def _print(*args: Any, **kwargs: Any) -> None:
                if "file" not in kwargs:
                    kwargs["file"] = out_buf

                return print(*args, **kwargs)

            eval_vars = {
                # Contextual info
                "self": self,
                "ctx": ctx,
                "bot": self.bot,
                "loop": self.bot.loop,
                "client": self.bot.client,
                "commands": self.bot.commands,
                "listeners": self.bot.listeners,
                "plugins": self.bot.plugins,
                "stdout": out_buf,
                # Convenience aliases
                "anjani": self.bot,
                "chat": ctx.chat,
                "context": ctx,
                "msg": ctx.msg,
                "message": ctx.msg,
                "db": self.bot.db,
                # Helper functions
                "send": send,
                "print": _print,
                # Built-in modules
                "inspect": inspect,
                "os": os,
                "re": re,
                "sys": sys,
                "traceback": traceback,
                # Third-party modules
                "pyrogram": pyrogram,
                # Custom modules
                "command": command,
                "plugin": plugin,
                "util": util,
            }

            try:
                with redirect_stdout(out_buf):
                    return "", await meval(code, globals(), **eval_vars)
            except Exception as e:  # skipcq: PYL-W0703
                # Find first traceback frame involving the snippet
                first_snip_idx = -1
                tb = traceback.extract_tb(e.__traceback__)
                for i, frame in enumerate(tb):
                    if frame.filename == "<string>" or frame.filename.endswith("ast.py"):
                        first_snip_idx = i
                        break

                # Re-raise exception if it wasn't caused by the snippet
                if first_snip_idx == -1:
                    raise e

                # Return formatted stripped traceback
                stripped_tb = tb[first_snip_idx:]
                formatted_tb = util.error.format_exception(e, tb=stripped_tb)
                return "⚠️ Error executing snippet\n\n", formatted_tb

        before = util.time.usec()
        prefix, result = await _eval()
        after = util.time.usec()

        # Always write result if no output has been collected thus far
        if not out_buf.getvalue() or result is not None:
            print(result, file=out_buf)

        el_us = after - before
        el_str = util.time.format_duration_us(el_us)

        out = out_buf.getvalue()
        # Strip only ONE final newline to compensate for our message formatting
        if out.endswith("\n"):
            out = out[:-1]

        respond_text = f"""{prefix}<b>In:</b>
<pre language="python">{escape(code)}</pre>

<b>Out:</b>
<pre language="python">{escape(out)}</pre>

Time: {el_str}"""
        if len(respond_text) > 4096:
            async with ctx.action(ChatAction.UPLOAD_DOCUMENT):
                with io.BytesIO(str.encode(out)) as out_file:
                    out_file.name = "eval.text"
                    await ctx.msg.reply_document(
                        document=out_file, caption=code, disable_notification=True
                    )

                return None

        await ctx.respond(
            respond_text,
            parse_mode=pyrogram.enums.parse_mode.ParseMode.HTML,
        )



# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import sys
from typing import IO, Any, Optional, Sequence, Tuple, Union

ProcessData = Union[str, bytes]
ProcessStream = Union[int, IO, None]


class FormatType:
    pass


StderrOnly = FormatType()


def get_venv_path() -> Optional[str]:
    """Returns the current venv's path prefix, or None if not running in a venv."""

    if hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix:
        return sys.prefix

    return None


async def _spawn_exec(
    cmdline: Sequence[ProcessData],
    in_data: Optional[bytes],
    stdout: ProcessStream,
    stderr: ProcessStream,
    **kwargs: Any,
) -> asyncio.subprocess.Process:
    stdin = asyncio.subprocess.PIPE if in_data else None
    return await asyncio.create_subprocess_exec(
        *cmdline, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs
    )


async def _spawn_shell(
    cmdline: ProcessData,
    in_data: Optional[bytes],
    stdout: ProcessStream,
    stderr: ProcessStream,
    **kwargs: Any,
) -> asyncio.subprocess.Process:
    stdin = asyncio.subprocess.PIPE if in_data else None
    return await asyncio.create_subprocess_shell(
        cmdline, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs
    )


async def _get_proc_output(
    proc: asyncio.subprocess.Process,
    in_data: Optional[bytes],
    timeout: Optional[int],
    text: Union[bool, FormatType],
) -> Tuple[ProcessData, ProcessData, Optional[int]]:
    stdout: Any
    stderr: Any
    try:
        stdout, stderr = await asyncio.wait_for(proc.communicate(in_data), timeout)
    except asyncio.TimeoutError:
        try:
            proc.kill()
        except ProcessLookupError:
            pass

        raise

    if text:
        if text is not StderrOnly and stdout is not None:
            stdout = stdout.decode(errors="replace").strip()

        if stderr is not None:
            stderr = stderr.decode(errors="replace").strip()

    return stdout, stderr, proc.returncode


async def run_command(
    *cmdline: ProcessData,
    in_data: Optional[bytes] = None,
    stdout: ProcessStream = asyncio.subprocess.PIPE,
    stderr: ProcessStream = asyncio.subprocess.STDOUT,
    timeout: Optional[int] = None,
    shell: bool = False,
    text: Union[bool, FormatType] = True,
    **kwargs: Any,
) -> Tuple[Any, Any, Optional[int]]:
    """Runs the given command (with optional input) using asyncio.subprocess."""

    if shell:
        proc = await _spawn_shell(cmdline[0], in_data, stdout, stderr, **kwargs)
    else:
        proc = await _spawn_exec(cmdline, in_data, stdout, stderr, **kwargs)

    return await _get_proc_output(proc, in_data, timeout, text)



""" Health Check plugin for @dAnjani_bot """

# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from datetime import datetime
from typing import ClassVar

from pyrogram.raw.functions.ping import Ping


from anjani import plugin


class Health(plugin.Plugin):
    name: ClassVar[str] = "HealthCheck"
    webhook_url: str
    interval: int

    # Private
    _run_check: bool = False
    __task: asyncio.Task[None]

    async def on_load(self) -> None:
        self.webhook_url = self.bot.config.HEALTH_CHECK_WEBHOOK_URL
        if not self.webhook_url:
            self.log.debug("Health Check Webhook URL is not set, disabling health check")
            self.bot.unload_plugin(self)
            return
        self._run_check = True
        self.interval = self.bot.config.HEALTH_CHECK_INTERVAL

    async def on_start(self, _: int) -> None:
        self.log.debug("Starting Health Check Push")
        self.__task = self.bot.loop.create_task(self.push_health())

    async def on_stop(self) -> None:
        self.log.debug("Stopping health check push")
        self._run_check = False
        self.__task.cancel()

    async def push_health(self) -> None:
        while self._run_check:
            try:
                await self.bot.http.get(
                    self.webhook_url,
                    params={"status": "up", "msg": "OK", "ping": await self.get_ping()},
                )
            except Exception as e:
                self.log.error(f"Error pushing health: {e}")

            await asyncio.sleep(self.interval)

    async def get_ping(self) -> float:
        time = datetime.now()
        await self.bot.client.invoke(Ping(ping_id=1))
        end = datetime.now()
        return (end - time).microseconds / 1000



"""Anjani plugin extender"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import inspect
from types import ModuleType as PluginType
from typing import TYPE_CHECKING, Any, Iterable, MutableMapping, Optional, Type

from anjani import custom_plugins, internal_plugins, plugin, plugins, util
from anjani.error import ExistingPluginError

from .anjani_mixin_base import MixinBase

if TYPE_CHECKING:
    from .anjani_bot import Anjani


class PluginExtender(MixinBase):
    # Initialized during instantiation
    plugins: MutableMapping[str, plugin.Plugin]

    def __init__(self: "Anjani", **kwargs: Any) -> None:
        # Initialize plugin map
        self.plugins = {}

        # Propagate initialization to other mixins
        super().__init__(**kwargs)

    def load_plugin(
        self: "Anjani", cls: Type[plugin.Plugin], *, comment: Optional[str] = None
    ) -> None:
        self.log.info(f"Loading {cls.format_desc(comment)}")

        if cls.name in self.plugins:
            old = type(self.plugins[cls.name])
            raise ExistingPluginError(old, cls)

        plug = cls(self)
        plug.comment = comment
        self.register_listeners(plug)
        self.register_commands(plug)
        self.plugins[cls.name] = plug

    def unload_plugin(self: "Anjani", plug: plugin.Plugin) -> None:
        cls = type(plug)
        self.log.info(f"Unloading {plug.format_desc(plug.comment)}")

        self.unregister_listeners(plug)
        self.unregister_commands(plug)
        del self.plugins[cls.name]

    def _load_all_from_metaplug(
        self: "Anjani", subplugins: Iterable[PluginType], *, comment: Optional[str] = None
    ) -> None:
        for plug in subplugins:
            for sym in dir(plug):
                cls = getattr(plug, sym)
                if inspect.isclass(cls) and issubclass(cls, plugin.Plugin):
                    name = cls.name.lower().replace(" ", "_")
                    if not self.config.is_plugin_disabled(name):
                        if not cls.disabled:
                            self.load_plugin(cls, comment=comment)
                        else:
                            # Add the disabled plugin from property to runtime config
                            self.config.PLUGIN_FLAG.append(f"disable_{name}_plugin")

    # noinspection PyTypeChecker,PyTypeChecker
    def load_all_plugins(self: "Anjani") -> None:
        self.log.info("Loading plugins")
        if self.config.is_flag_active("enable_internal_plugin"):
            self._load_all_from_metaplug(internal_plugins.subplugins, comment="internal")
        self._load_all_from_metaplug(custom_plugins.subplugins, comment="custom")
        self._load_all_from_metaplug(plugins.subplugins)
        self.log.info("All plugins loaded.")

    def unload_all_plugins(self: "Anjani") -> None:
        self.log.info("Unloading plugins...")

        # Can't modify while iterating, so collect a list first
        for plug in list(self.plugins.values()):
            self.unload_plugin(plug)

        self.log.info("All plugins unloaded.")

    async def reload_plugin_pkg(self: "Anjani") -> None:
        self.log.info("Reloading base plugin class...")
        await util.run_sync(importlib.reload, plugin)

        self.log.info("Reloading master plugin...")
        await util.run_sync(importlib.reload, plugins)

        self.log.info("Reloading custom master module...")
        await util.run_sync(importlib.reload, custom_plugins)



"""User language setting"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from functools import partial
from typing import Any, ClassVar, MutableMapping, Optional

from pymongo.errors import PyMongoError
from pyrogram import emoji
from pyrogram.enums.chat_type import ChatType
from pyrogram.errors import MessageNotModified
from pyrogram.types import (
    CallbackQuery,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
)

from anjani import command, filters, listener, plugin, util

LANG_FLAG = {
    "en": f"{emoji.FLAG_UNITED_STATES} English",
    "id": f"{emoji.FLAG_INDONESIA} Indonesia",
}


class Language(plugin.Plugin):
    """Bot language plugin"""

    name: ClassVar[str] = "Language"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    _db_stream: asyncio.Task[None]

    def _start_db_stream(self) -> None:
        # Cancel previous task if exists and it's not done
        try:
            if not self._db_stream.done():
                self._db_stream.cancel()
        except AttributeError:
            pass

        self._db_stream = self.bot.loop.create_task(self.db_stream())
        self._db_stream.add_done_callback(
            partial(self.bot.loop.call_soon_threadsafe, self._db_stream_callback)
        )

    def _db_stream_callback(self, future: asyncio.Future) -> None:
        try:
            future.result()
        except asyncio.CancelledError:
            pass
        except PyMongoError as e:
            self.log.error("MongoDB error:", exc_info=e)
            self._start_db_stream()

    async def db_stream(self) -> None:
        async with self.db.watch(full_document="updateLookup") as cursor:
            async for change in cursor:
                document = change["fullDocument"]
                self.bot.chats_languages[document["chat_id"]] = document["language"]

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("LANGUAGE")
        self._start_db_stream()

    async def on_stop(self) -> None:
        self._db_stream.cancel()

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        language = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: language} if language else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @listener.filters(filters.regex(r"set_lang_(.*)"))
    async def on_callback_query(self, query: CallbackQuery) -> None:
        """Set language query."""
        lang_match = query.matches[0].group(1)
        chat = query.message.chat

        # Check admin rights
        if chat.type != ChatType.PRIVATE:
            user = await chat.get_member(query.from_user.id)
            if not user.privileges or not user.privileges.can_change_info:
                await query.answer(await self.text(chat.id, "error-no-rights"))
                return

        lang = LANG_FLAG.get(lang_match)
        if not lang:
            await query.edit_message_text(await self.text(chat.id, "language-code-error"))
            return

        await self.switch_lang(chat.id, lang_match)
        try:
            await query.answer()
            await query.edit_message_text(
                text=await self.text(chat.id, "language-set-succes", lang),
            )
        except MessageNotModified:
            await query.answer(
                await self.text(chat.id, "language-set-succes", lang), show_alert=True
            )
            await query.message.delete()

    async def switch_lang(self, chat_id: int, language: str) -> None:
        """Change chat language setting."""
        await self.db.update_one(
            {"chat_id": int(chat_id)},
            {"$set": {"language": language}},
            upsert=True,
        )

    @command.filters(aliases=["lang", "language"])
    async def cmd_setlang(self, ctx: command.Context) -> Optional[str]:
        """Set user/chat language."""
        chat = ctx.chat

        # Check admin rights
        if chat.type != ChatType.PRIVATE:
            user = await chat.get_member(ctx.msg.from_user.id)
            if not user.privileges or not user.privileges.can_change_info:
                return await self.text(chat.id, "error-no-rights")

        if ctx.input:
            lang = ctx.input.lower()
            if lang in self.bot.languages:
                await asyncio.gather(
                    self.switch_lang(chat.id, lang),
                    ctx.respond(await self.text(chat.id, "language-set-succes", LANG_FLAG[lang])),
                )
            else:
                return await self.text(chat.id, "language-invalid", list(self.bot.languages.keys()))
        else:
            chat_name = chat.first_name or chat.title
            lang = LANG_FLAG[self.bot.chats_languages.get(chat.id, "en")]
            keyboard = []
            temp = []

            for count, i in enumerate(self.bot.languages.keys(), start=1):
                temp.append(InlineKeyboardButton(LANG_FLAG[i], callback_data=f"set_lang_{i}"))
                if count % 2 == 0:
                    keyboard.append(temp)
                    temp = []
                if count == len(self.bot.languages):
                    keyboard.append(temp)
            await ctx.respond(
                await self.text(chat.id, "current-language", chat_name, lang),
                reply_markup=InlineKeyboardMarkup(keyboard),
            )

        return None



"""Anjani database cursor core"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import inspect
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Coroutine,
    Deque,
    Generic,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.cursor import _QUERY_OPTIONS, Cursor, RawBatchCursor
from pymongo.typings import _Address, _DocumentType

from anjani import util

from .base import AsyncBase
from .errors import InvalidOperation

if TYPE_CHECKING:
    from .collection import AsyncCollection
    from .command_cursor import CommandCursor, _LatentCursor


class AsyncCursorBase(AsyncBase, Generic[_DocumentType]):
    """Base class for Cursor AsyncIOMongoDB instances

    *DEPRECATED* methods are removed in this class.

    :meth:`~each()` is removed because we can iterate directly this class,
    And we now have :meth:`~to_list()` so yeah kinda useless
    """

    collection: Optional[Union["AsyncCollection[_DocumentType]", Collection[_DocumentType]]]
    dispatch: Union[
        "_LatentCursor[_DocumentType]",
        "CommandCursor[_DocumentType]",
        Cursor[_DocumentType],
        RawBatchCursor[_DocumentType],
    ]
    loop: asyncio.AbstractEventLoop

    def __init__(
        self,
        cursor: Union[
            "_LatentCursor[_DocumentType]",
            "CommandCursor[_DocumentType]",
            Cursor[_DocumentType],
            RawBatchCursor[_DocumentType],
        ],
        collection: "Optional[AsyncCollection[_DocumentType]]" = None,
    ) -> None:
        super().__init__(cursor)

        if collection:
            self.collection = collection
        else:
            self.collection = cursor.collection
        self.started = False
        self.closed = False

        self.loop = asyncio.get_event_loop()

    async def __aenter__(self) -> "AsyncCursorBase":
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.dispatch:
            await self.close()

    def __aiter__(self) -> "AsyncCursorBase":
        return self

    async def __anext__(self) -> Mapping[str, Any]:
        return await self.next()

    def _buffer_size(self) -> int:
        return len(self._data())

    def _query_flags(self) -> int:
        raise NotImplementedError

    def _data(self) -> Deque[Any]:
        raise NotImplementedError

    def _killed(self) -> bool:
        raise NotImplementedError

    def _get_more(self) -> Coroutine[Any, Any, int]:
        if not self.alive:
            raise InvalidOperation(
                "Can't call get_more() on a AsyncCursor that has been" " exhausted or killed."
            )

        self.started = True
        return self._refresh()

    def _to_list(
        self,
        length: Optional[int],
        the_list: List[Mapping[str, Any]],
        future: asyncio.Future[List[Mapping[str, Any]]],
        get_more_future: asyncio.Future[int],
    ) -> None:
        # get_more_future is the result of self._get_more().
        # future will be the result of the user's to_list() call.
        try:
            result = get_more_future.result()
            # Return early if the task was cancelled.
            if future.done():
                return

            if length is None:
                n = result
            else:
                n = min(length - len(the_list), result)

            i = 0
            while i < n:
                the_list.append(self._data().popleft())
                i += 1

            reached_length = length is not None and len(the_list) >= length
            if reached_length or not self.alive:
                future.set_result(the_list)
            else:
                new_future = self.loop.create_task(self._get_more())
                new_future.add_done_callback(
                    partial(self.loop.call_soon_threadsafe, self._to_list, length, the_list, future)
                )
        except Exception as exc:  # skipcq: PYL-W0703
            if not future.done():
                future.set_exception(exc)

    async def _refresh(self) -> int:
        return await util.run_sync(self.dispatch._refresh)  # skipcq: PYL-W0212

    def batch_size(self, batch_size: int) -> "AsyncCursorBase":
        self.dispatch.batch_size(batch_size)
        return self

    async def close(self) -> None:
        if not self.closed:
            self.closed = True
            await util.run_sync(self.dispatch.close)

    async def next(self) -> Any:
        if self.alive and (self._buffer_size() or await self._get_more()):
            return await util.run_sync(next, self.dispatch)
        raise StopAsyncIteration

    def to_list(self, length: Optional[int] = None) -> asyncio.Future[List[Mapping[str, Any]]]:
        if length is not None and length < 0:
            raise ValueError("length must be non-negative")

        if self._query_flags() & _QUERY_OPTIONS["tailable_cursor"]:
            raise InvalidOperation("Can't call to_list on tailable cursor")

        future = self.loop.create_future()
        the_list: List[Mapping[str, Any]] = []

        if not self.alive:
            future.set_result(the_list)
            return future

        # Ignored the type since some commands are called from command_cursor
        get_more_future: Union[asyncio.Future, asyncio.Task] = self._get_more()  # type: ignore
        if inspect.iscoroutine(get_more_future):
            get_more_future = self.loop.create_task(get_more_future)

        get_more_future.add_done_callback(
            partial(self.loop.call_soon_threadsafe, self._to_list, length, the_list, future)
        )

        return future

    @property
    def address(self) -> Optional[Union[Tuple[str, int], _Address]]:
        return self.dispatch.address

    @property
    def alive(self) -> bool:
        if not self.dispatch:
            return True
        return self.dispatch.alive

    @property
    def cursor_id(self) -> Optional[int]:
        return self.dispatch.cursor_id

    @property
    def session(self) -> Optional[ClientSession]:
        return self.dispatch.session
